#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright 2014 Brno University of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# File:   check_namelist.py
# Author: Peter Hostačný, xhosta03@stud.fit.vutbr.cz
# Date:   25.2.2015

import re
import sys
from os.path import isfile
import getopt


################## GLOBAL CONSTANTS ###################
bin_numbers = False


def print_help():
	''' Print help. '''

	print('''
Author:        Peter Hostacny (xhosta03@stud.fit.vutbr.cz)
Description:   Script controls validity of namelists for fsa_build. There is no guaranty of complete validity, but it can find a lot of bugs. Namelist should be in ASCII format with normal/binary numbers (argument -b for binary numbers).

REQUIRE: Python 3

Examples of correct line:
Entity in namelist|15455;18752
Alle Menschen werden Br&#xC3BC;der|2067522
Some entity   |4583

Examples of incorrect line:
Entity on line with missing pipe 2547
Missing number after semicolon|1542;
|1545
Missing number|

Usage:   ./check_namelist.py [OPTIONS]

Options:
  -h, --help           -> Show this help message and exit.
  -b                   -> Namelist with binary numbers. (not implemented)
''')

def process_arguments(args):
	''' Parse arguments from the command line and evaluate them. '''

	global bin_numbers

	try:
		opts, remaining = getopt.getopt(args[1:], "hb", ["help"])
	except getopt.GetoptError as err:
		print(err)
		sys.exit(1)

	for opt, arg in opts:
		if opt in ("-h", "--help"):
			print_help()
			sys.exit(0)
		if opt in ("-b"):
			bin_numbers = True
		
	for namelist in remaining:
		if not (isfile(namelist)):
			print("File '" + namelist + "' doesn't exist.")
			sys.exit(2)

	return remaining


def check_it(lines):
	''' argument: list of lines  '''
	result = ""
	count = 0
	max_chars = 0
	
	line_pattern = re.compile("^[^|]+\|[0-9]+(;[0-9]+)*$")	

	for line in lines:
		if len(line) > max_chars:
			max_chars = len(line)
		match = line_pattern.match(line)
		count += 1
		if not match:
			result += str(count) + ": " + line + "    # Bad format.\n"

		try:
			line.encode("ascii")
		except UnicodeEncodeError:
			result += str(count) + ": " + line +  "    # There is a unicode character in namelist.\n"

	print("max. lenght of line: " + str(max_chars))
	return result




if __name__ == '__main__':

	result = ""
	print()

	namelists = process_arguments(sys.argv)
	if len(namelists) != 0:
		for namelist in namelists:
			with open(namelist) as input_file:
				result = check_it(input_file.read().splitlines())
			print("===  " + namelist + "  ===")
			if len(result) == 0:
				print("O.K.")
			else:
				print(result)
			print()

