#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2015 Brno University of Technology

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Author: Matej Magdolen, xmagdo00@stud.fit.vutbr.cz
# Author: Jan Cerny, xcerny62@stud.fit.vutbr.cz
# Author: Lubomir Otrusina, iotrusina@fit.vutbr.cz
#
# Description: Reads text from the standard input or from a file, finds
#              knowledge base entities and dates, disambiguates entities,
#              resolves corefernces.

import sys
import re
import dates
import argparse
import figa.sources.marker as figa
import ner_knowledge_base
import os
import unicodedata

reload(sys)
sys.setdefaultencoding("utf-8")

# pronouns along with their type they refer to
PRONOUNS = {"he" : "M",
           "him" : "M",
       "himself" : "M",
           "his" : "M",
           "she" : "F",
           "her" : "F",
          "hers" : "F",
       "herself" : "F",
           "who" : "MF",
          "whom" : "MF",
         "whose" : "MF",
          "here" : "L",
         "there" : "L",
         "where" : "L"}

# a list with the most famous artworks (don't need the author to be mentioned in the input text)
MOST_FAMOUS_ARTWORKS = [line.strip().lower() for line in open(os.path.abspath(os.path.join(ner_knowledge_base.SCRIPT_DIR, "./most_famous_artworks")))]

def remove_accent(_string):
    """ Removes accents from a string. For example, Eduard Ovčáček -> Eduard Ovcacek. """

    nkfd_form = unicodedata.normalize('NFKD', unicode(_string))
    return str("".join([c for c in nkfd_form if not unicodedata.combining(c)]))

def ncr2unicode(_string):
    """ Translates hexadecimal NCRs to an 8-bit string version of the Unicode. For example, &#x957F;&#x5EA6; (长度) -> \\x95\\x7F\\x5E\\xA6. """

    _string = re.sub("&#x([A-F0-9]{2});", "\\x\g<1>", _string)
    _string = re.sub("&#x([A-F0-9]{2})([A-F0-9]{2});", "\\x\g<1>\\x\g<2>", _string)
    _string = re.sub("&#x([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2});", "\\x\g<1>\\x\g<2>\\x\g<3>", _string)
    _string = re.sub("&#x([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2});", "\\x\g<1>\\x\g<2>\\x\g<3>\\x\g<4>", _string)
    _string = re.sub("&#x([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2});", "\\x\g<1>\\x\g<2>\\x\g<3>\\x\g<4>\\x\g<5>", _string)
    _string = re.sub("&#x([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2})([A-F0-9]{2});", "\\x\g<1>\\x\g<2>\\x\g<3>\\x\g<4>\\x\g<5>\\x\g<6>", _string)
    return eval("\"" + _string + "\"")

class EntityRegister:
    """ A class containing the index of all disambiguated entities. """

    def __init__(self):
        self.id2entity = {}
        self.entity2id = {}

    def insert_entity(self, _entity, _id):
        """ Insterts a preferred sense for a given entity into to the entity register. """

        if _entity in self.entity2id:
            sense = self.entity2id[_entity]
            self.id2entity[sense].discard(_entity)
        self.entity2id[_entity] = _id
        if _id not in self.id2entity:
            self.id2entity[_id] = set()
        self.id2entity[_id].add(_entity)

    def __str__(self):
        return str(self.id2entity)

class Entity:
    """ A text entity referring to a knowledge base item. """

    def __init__(self, entity_str, kb, input_string, original_input, register):
        """
        Creates an entity by parsing a line of figa output from entity_str.
        Entity will be referring to an item of the knowledge base kb.

        entity_str - a line from the figa output
        kb - Knowledge Base
        input_string - input string
        original_input - original input string
        register - entity register
        """

        self.next_to_same_type = False
        self.display_score = False
        self.poorly_disambiguated = True
        self.is_coreference = False

        self.preferred_sense = None
        self.next_word_begin = None
        self.next_word_end = None
        self.previous_word = None
        self.before_previous = None
        self.candidates = []
        self.score = []
        self.coreferences = set()

        self.input_string = input_string
        self.original_input = original_input
        self.kb = kb
        self.register = register

        # parsing entity attributes
        entity_attributes = entity_str.split('\t')

        # getting possible senses (sense 0 marks a coreference)
        self.senses = set([int(s) for s in entity_attributes[0].split(';') if s != '0'])

        # start offset is indexed differently from figa
        self.start_offset = int(entity_attributes[1]) - 1
        self.end_offset = int(entity_attributes[2])

        # the source text of the entity
        self.source = ncr2unicode(entity_attributes[3])

        # possible coreferences - people whose names are supersets of an entity
        self.partial_match_senses = self.kb.people_named(remove_accent(self.source).lower())

    def set_preferred_sense(self, _sense):
        self.preferred_sense = _sense

        if not isinstance(_sense, Entity):
            self.register.insert_entity(self, _sense)

    def has_preferred_sense(self):
        return self.preferred_sense

    def get_preferred_sense(self):
        if isinstance(self.preferred_sense, Entity):
            return self.preferred_sense.preferred_sense
        else:
            return self.preferred_sense

    def get_preferred_entity(self):
        if not isinstance(self.preferred_sense, Entity):
            return self
        else:
            return self.preferred_sense

    def disambiguate_without_context(self):
        """ Chooses the correct sense of the entity as the preferred one (without context). """

        # we don't resolve coreference in this step
        if self.source.lower() in PRONOUNS or self.partial_match_senses:
            self.is_coreference = True
            return

        # locations cannot end with 's
        self.senses = [s for s in self.senses if not (self.kb.get_ent_type(s) == "location" and self.right_context("'s"))]

        # locations cannot start with The
        self.senses = [s for s in self.senses if not (self.kb.get_ent_type(s) == "location" and self.source.startswith("The "))]

        # only locations can have preposition "into"
        self.senses = [s for s in self.senses if not (self.kb.get_ent_type(s) != "location" and self.left_context(" into "))]

        # if candidates contain any location, excludes all nationalities, events and artworks
        for sense in self.senses:
            ent_type = self.kb.get_ent_type(sense)
            if ent_type == "location":
                self.senses = [s for s in self.senses if self.kb.get_ent_type(s) not in ["nationality", "event", "artwork"]]
                break

        # if candidates contain any artist, excludes all groups
        for sense in self.senses:
            ent_type = self.kb.get_ent_type(sense)
            if ent_type == "artist":
                self.senses = [s for s in self.senses if self.kb.get_ent_type(s) not in ["group"]]
                break

        # coreferences by a name are not resolved without a context
        self.senses = set(self.senses)
        self.candidates = [s for s in self.senses]

        if not self.candidates:
            return
        elif len(self.candidates) == 1:
            self.set_preferred_sense(self.candidates[0])
            self.poorly_disambiguated = False

        # the entity has to be disambiguated
        if not self.has_preferred_sense():
            for i in self.candidates:
                static_score = self.kb.get_score(i)
                self.score.append(static_score)

            self.set_preferred_sense(self.candidates[self.score.index(max(self.score))])

    def disambiguate_with_context(self, context):
        """ Chooses the correct sense of the entity as the preferred one (with context). """

        # we don't resolve coreference in this step
        if self.is_coreference or not self.candidates:
            return

        # recomputing paragraph offset
        context.recompute_paragraph_offset(self.start_offset)

        # the entity has to be disambiguated
        self.score = []
        for i in self.candidates:
            ent_type = self.kb.get_ent_type(i)
            static_score = self.kb.get_score(i)
            context_score = 0
            if ent_type == 'location':
                context_score = context.country_percentile(self.kb.get_data_for(i, "COUNTRY"))
            elif ent_type in ['person', 'artist']:
                context_score = context.person_percentile(i)
            if context_score > 0:
                self.poorly_disambiguated = False
            self.score.append(static_score + context_score)
        self.set_preferred_sense(self.candidates[self.score.index(max(self.score))])

    def resolve_pronoun_coreference(self, context):
        """ Resolves a pronoun coreference using context. """

        if self.source == "There" and (self.right_context(" is ") or self.right_context(" are ") or self.right_context(" was ") or self.right_context(" were ") or self.right_context(" has ") or self.right_context(" have ") or self.right_context(" had ")):
            return

        pronoun_type = PRONOUNS[self.source.lower()]
        begin_of_paragraph = self.begin_of_paragraph(context)

        if pronoun_type == 'M':
            if self.source == "him" and context.before_last_person and context.before_last_person.start_offset >= begin_of_paragraph:
                self.set_preferred_sense(context.before_last_person.get_preferred_entity())
            else:
                if context.last_unknown_gender:
                    context.before_last_person = context.last_male
                    context.last_male = context.last_unknown_gender
                    context.last_unknown_gender = None
                if context.last_male and context.last_male.start_offset >= begin_of_paragraph:
                    self.set_preferred_sense(context.last_male.get_preferred_entity())
        elif pronoun_type == 'F':
            if context.last_unknown_gender:
                context.last_female = context.last_unknown_gender
                context.last_unknown_gender = None
            if context.last_female and context.last_female.start_offset >= begin_of_paragraph:
                self.set_preferred_sense(context.last_female.get_preferred_entity())
        elif pronoun_type == 'N':
            if context.last_thing and context.last_thing.start_offset >= begin_of_paragraph:
                self.set_preferred_sense(context.last_thing.get_preferred_entity())
        elif pronoun_type == 'MF':
            if context.last_person and context.last_person.start_offset >= begin_of_paragraph:
                self.set_preferred_sense(context.last_person.get_preferred_entity())
        elif pronoun_type == 'L':
            if context.last_location and context.last_location.start_offset >= begin_of_paragraph:
                self.set_preferred_sense(context.last_location.get_preferred_entity())

    def begin_of_paragraph(self, context):
        """ Returns the start offset of the paragraph the entity belongs to. """

        x = 0
        for beg_of_par in context.paragraphs:
            if beg_of_par > self.start_offset:
                 break
            x = beg_of_par
        return x

    def __str__(self):
        """ Converts an entity into an output format. """

        result = str(self.start_offset) + "\t" + str(self.end_offset) + "\t"
        if self.is_coreference:
            result += "coref"
        else:
            result += "kb"
        result += "\t" + self.original_input[self.start_offset:self.end_offset] + "\t"
        if self.display_score and self.candidates:
            candidates_str = []
            i = 0
            for cand in self.candidates:
                candidates_str.append(str(cand))
                if i < len(self.score):
                    candidates_str[-1] += " " + str(self.score[i])
                i += 1
            result += ";".join(candidates_str)
        elif self.has_preferred_sense():
            result += str(self.get_preferred_sense())
        else:
            if self.is_coreference:
                senses_list = sorted(self.partial_match_senses)
            else:
                senses_list = sorted(self.senses)
            for i in senses_list:
                result += str(i)
                if i != senses_list[-1]:
                    result += ';'
        return result

    def right_context(self, right):
        length = len(right)
        text = self.input_string.decode("utf8", "replace").encode('ascii', 'replace')
        if self.end_offset + length > len(text):
            return False
        return text[self.end_offset:self.end_offset + length] == right

    def left_context(self, left):
        length = len(left)
        text = self.input_string.decode("utf8", "replace").encode('ascii', 'replace')
        if self.start_offset - length < 0:
            return False
        return text[self.start_offset - length:self.start_offset] == left

class Context:
    """ Information about a context of a processed text. """

    def __init__(self, entities, kb, paragraphs):
        """ Prepares the context from the list of entities disambiguated without the context. """

        self.entities = entities
        self.kb = kb
        self.paragraphs = paragraphs

        # max score for each candidate
        self.people_max_scores = {}
        # number of locations in each country mentioned in each paragraph (+ sorted version)
        self.countries = {}
        self.countries_sorted = {}
        # people mentioned by full name
        self.people = set()
        # list of nationalities mentioned in each paragraph (may contain duplicates)
        self.people_nationalities = {}
        # list of dates mentioned in each paragraph (may contain duplicates)
        self.people_dates = {}
        # initializing pronoun variables
        self.init_pronouns()
        # initializing the paragraph index
        self.paragraph_index = 0

        # initializing index variables
        par_index = 0
        ent_index = 0

        # adding one artificial paragraph, otherwise self.paragraphs[par_index + 1] will fail
        self.paragraphs.append(sys.maxint)

        # computing statistics for each paragraph
        for par in self.paragraphs:
            self.countries[par] = {}
            self.people_nationalities[par] = []
            self.people_dates[par] = []
            while (ent_index < len(entities) and entities[ent_index].start_offset < self.paragraphs[par_index + 1]):
                ent = self.entities[ent_index]
                if isinstance(ent, Entity) and ent.has_preferred_sense():
                    ent_type = ent.kb.get_ent_type(ent.get_preferred_sense())
                    if ent_type == 'location':
                        country = ent.kb.get_data_for(ent.get_preferred_sense(), "COUNTRY")
                        if country != "":
                            if country not in self.countries[par]:
                                self.countries[par][country] = 0
                            self.countries[par][country] += 1
                    elif ent_type == 'nationality':
                        self.people_nationalities[par].append(ent.kb.get_nationalities(ent.get_preferred_sense()))
                elif isinstance(ent, dates.Date):
                    # removing months and days with zeros (only years will remain)
                    if ent.class_type == ent.Type.DATE:
                        self.people_dates[par].append(re.sub("-(00)+", "", str(ent.iso8601)))
                    elif ent.class_type == ent.Type.INTERVAL:
                        self.people_dates[par].append(re.sub("-(00)+", "", str(ent.date_from)))
                        self.people_dates[par].append(re.sub("-(00)+", "", str(ent.date_to)))
                ent_index += 1
            par_index += 1

        # removing the artificial paragraph
        self.paragraphs.pop()

    def recompute_paragraph_offset(self, start_offset):
        """
        Recomputes paragraph offset, if the entity at the start_offset belongs
        to the different paragraph.
        """

        # we are in the last paragraph -> no change in paragraph offsets
        if self.paragraph_index + 1 >= len(self.paragraphs):
            return
        # the entity belongs to the current paragraph -> no change in paragraph offsets
        elif start_offset >= self.paragraphs[self.paragraph_index] and start_offset < self.paragraphs[self.paragraph_index + 1]:
            return
        # the entity belongs to the different paragraph -> we have to recompute paragraph offsets
        else:
            par_i = self.paragraph_index
            while par_i + 1 < len(self.paragraphs) and self.paragraphs[par_i + 1] <= start_offset:
                par_i += 1
            self.paragraph_index = par_i

    def update(self, entity):
        """ Updates context (last person, last male, last female, last location etc.). """

        # keep the last entity of each pronoun type for pronoun coreference resolution
        ent_type = self.kb.get_ent_type(entity.get_preferred_sense())

        if ent_type in ['person', 'artist']:
            self.before_last_person = self.last_person
            self.last_person = entity
            gender = self.kb.get_data_for(entity.get_preferred_sense(), "GENDER")
            if gender == 'M':
                self.last_male = entity
                self.last_unknown_gender = None
            elif gender == 'F':
                self.last_female = entity
                self.last_unknown_gender = None
            else:
                self.last_unknown_gender = entity
        elif ent_type == "location":
            self.last_location = entity
        else:
            self.last_thing = entity

    def person_percentile(self, candidate):
        """
        Returns a percentile of references to a candidate person from
        knowledge base amongst other people.
        """

        # computing people_nation_score
        people_nationality_score = 0
        # getting the nationality for a given person
        person_nationality = self.kb.get_nationalities(candidate)
        for nat in self.people_nationalities[self.paragraphs[self.paragraph_index]]:
            if nat & person_nationality:
                # the person has the same nationality like other persons in this paragraph
                people_nationality_score += 1
        if self.people_nationalities[self.paragraphs[self.paragraph_index]]:
           # normalizing people_nationality_score
           people_nationality_score = people_nationality_score * 100 / len(self.people_nationalities[self.paragraphs[self.paragraph_index]])

        # computing people_date_score
        people_date_score = 0
        # getting the dates for a given person
        person_dates = self.kb.get_dates(candidate)
        for context_date in self.people_dates[self.paragraphs[self.paragraph_index]]:
            for person_date in person_dates:
                if context_date.find(person_date) > -1 or person_date.find(context_date) > -1:
                    # the person has the date mentioned in this paragraph
                    people_date_score += 1
        if self.people_dates[self.paragraphs[self.paragraph_index]]:
           # normalizing people_date_score
            people_date_score = people_date_score * 100 / len(self.people_dates[self.paragraphs[self.paragraph_index]])

        # summing up the scores
        result = (people_nationality_score + people_date_score) / 2

        # storing new max score
        if candidate in self.people_max_scores and result > self.people_max_scores[candidate]:
                self.people_max_scores[candidate] = result
        else:
            self.people_max_scores[candidate] = result

        return result

    def country_percentile(self, country):
        """ Returns a percentile of a number of location belonging to a country identified by code. """

        if country not in self.countries[self.paragraphs[self.paragraph_index]]:
            # we don't have any statistics for a given country
            return 0
        else:
            # computing and normalizing score for a given country
            return self.countries[self.paragraphs[self.paragraph_index]][country] * 100 / sum(self.countries[self.paragraphs[self.paragraph_index]].values())

    def init_pronouns(self):
        """ Initializes pronoun variables. """

        self.before_last_person = None
        self.last_person = None
        self.last_male = None
        self.last_female = None
        self.last_unknown_gender = None
        self.last_thing = None
        self.last_location = None

def offsets_of_paragraphs(input_string):
    """ Returns a list of starting offsets of each paragraph in input_string. """

    result = [0]
    result.extend([par_match.end() for par_match in re.finditer("\n+", input_string)])
    return result

def find_proper_nouns(input_string):
    """ Returns a list of proper nouns. """

    result = []
    proper_noun_regex = re.compile(r"[A-Z][A-Za-z\'\-]*([A-Z][A-Za-z\'\-]*|the|upon| )*[A-Z][A-Za-z\'\-]*")
    for pn in re.finditer(proper_noun_regex, input_string):
        fields = pn.group(0).split()
        if fields[0] not in ["On", "At", "But", "In", "Mayor"]:
            result.append((pn.start(), pn.end()))
    return result

def fix_poor_disambiguation(entities, context):
    """ Fixes the entity sense if poorly_disambiguated is set to True. """

    strong_entities = {}
    entities = [e for e in entities if isinstance(e, Entity) and not e.is_coreference]

    for e in entities:
        if not e.poorly_disambiguated:
            if e.source not in strong_entities:
                strong_entities[e.source] = []
            strong_entities[e.source].append(e.get_preferred_entity())

    for e in entities:
        if e.poorly_disambiguated:
            if e.source in strong_entities:
                e.set_preferred_sense(get_nearest_entity(e, strong_entities[e.source]))
                e.poorly_disambiguated = False

def resolve_coreferences(entities, context, print_all, register):
    """ Resolves coreferences. """

    for e in entities:
        # adding propriate candidates into people set
        if not e.is_coreference and e.has_preferred_sense():
            ent_type = e.kb.get_ent_type(e.get_preferred_sense())
            if ent_type in ['person', 'artist']:
                context.people.add(e.get_preferred_sense())

    for e in entities:
        # resolving pronoun coreference
        if e.source.lower() in PRONOUNS:
            e.resolve_pronoun_coreference(context)
        elif e.is_coreference:
            # coreferences by a name to people out of context are discarded
            if not print_all:
                e.partial_match_senses = set([s for s in e.partial_match_senses if s in context.people])
                if e.partial_match_senses:
                    # choosing the candidate with the highest confidence score
                    sense = sorted(list(e.partial_match_senses), key=lambda candidate: -context.kb.get_score(candidate))[0]
                    candidates = list(register.id2entity[sense])
                    if not e.source.lower().startswith("the"):
                        # each candidate has to contain the text of a given entity
                        candidates = (c for c in candidates if e.source in c.source)
                    # choosing the nearest predecessor candidate for a coreference
                    entity = get_nearest_predecessor(e, candidates)
                    if entity:
                        e.set_preferred_sense(entity)
                elif e.senses:
                    e.is_coreference = False
                    e.disambiguate_without_context()
                    e.disambiguate_with_context(context)
        if e.has_preferred_sense():
            context.update(e)

def get_nearest_predecessor(_entity, _candidates):
    """ Returns the nearest predecessor for a given entity from a given list of candidates. """

    # sorting candidates according to the distance from a given entity
    candidates = sorted(_candidates, key=lambda candidate: _entity.start_offset - candidate.start_offset)
    for candidate in candidates:
        if _entity.start_offset - candidate.start_offset > 0:
            return candidate

def get_nearest_entity(_entity, _candidates):
    """ Returns the nearest entity for a given entity from a given list of candidates. """

    # sorting candidates according to the distance from a given entity
    candidates = sorted(_candidates, key=lambda candidate: abs(_entity.start_offset - candidate.start_offset))
    return candidates[0].preferred_sense

dictionary = None
lang_file = None

def recognize(kb, input_string, print_all=False, print_result=True, print_score=False, lowercase=False, remove=False):
    """
    Prints a list of entities found in input_string.

    kb - a knowledge base
    print_all - if false, all entities are disambiguated
    print_result - if True, the result is both returned as a list of entities and printed to stdout; otherwise, it is only returned
    print_score - similar to print_all, but also prints the score for each entity alternative
    lowercase - the input string is lowercased
    remove - removes accent from the input string
    """

    # original_input is intended for read-only operations
    original_input = input_string.decode("utf_8", "replace")

    # replacing some useless characters with a space character
    input_string = re.sub("[;\x01-\x08\x0e-\x1f\x0c\x7f]", " ", input_string)

    # running with parametr --remove_accent
    if remove:
        input_string = remove_accent(input_string)

    global dictionary
    global lang_file

    if not dictionary:
        dictionary = figa.myList()
        lang_file = None

        lower = ""
        if lowercase:
            lower = "-lower"

        dictionary.insert(os.path.dirname(os.path.realpath(__file__)) + "/figa/automata" + lower + ".fsa")
    seek_names = figa.marker(dictionary, lang_file)

    entities = []

    # getting data from figa
    if lowercase:
        output = seek_names.lookup_string(input_string.lower())
    else:
        output = seek_names.lookup_string(input_string)

    # a set of all possible senses
    global_senses = set()

    # creating entity register
    register = EntityRegister()

    # processing figa output and creating Entity objects
    for line in output.split("\n")[:-1]:
        e = Entity(line, kb, input_string, original_input, register)
        global_senses.update(e.senses)
        entities.append(e)

    # retaining only possible coreferences for each entity
    for e in entities:
        e.partial_match_senses = e.partial_match_senses & global_senses

    # searches for dates and intervals in the input
    dates_and_intervals = dates.find_dates(input_string)

    # resolving overlapping dates and entities
    entity_offsets = set()
    for e in entities:
        entity_offsets.update(set(range(e.start_offset, e.end_offset + 1)))
    dates_and_intervals = [d for d in dates_and_intervals if set(range(d.start_offset, d.end_offset + 1)) & entity_offsets == set()]

    # removing shorter entity from overlapping entities
    # figa should always return the longest match first
    entity_offsets = set()
    new_entities = []
    for e in entities:
        current_offset = set(range(e.start_offset, e.end_offset + 1))
        if current_offset & entity_offsets == set():
            entity_offsets.update(current_offset)
            new_entities.append(e)
    entities = new_entities

    # removing entities without any sense
    entities = [e for e in entities if e.senses or e.partial_match_senses or e.source.lower() in PRONOUNS]

    # merges entities with dates
    entities_and_dates = []
    entities_and_dates.extend(dates_and_intervals)
    entities_and_dates.extend(entities)

    # sorts entities and dates according to their start offsets
    entities_and_dates.sort(key=lambda ent : ent.start_offset)

    # disambiguates without context
    [e.disambiguate_without_context() for e in entities]
    paragraphs = offsets_of_paragraphs(original_input)
    context = Context(entities_and_dates, kb, paragraphs)
    # disambiguates with context
    [e.disambiguate_with_context(context) for e in entities]
    fix_poor_disambiguation(entities, context)
    context = Context(entities_and_dates, kb, paragraphs)

    # creating a set of global artist names
    global_artist_names = set()
    for e in entities:
        if e.has_preferred_sense():
            preferred_sense = e.get_preferred_sense()
            if e.kb.get_ent_type(preferred_sense) == "artist":
                global_artist_names.add(e.kb.get_data_for(preferred_sense, "DISPLAY TERM"))
                global_artist_names.add(e.kb.get_data_for(preferred_sense, "PREFERRED TERM"))
                global_artist_names.update(set(e.kb.get_data_for(preferred_sense, "OTHER TERM").split("|")))

    # removing artworks created by artist not presented in the document
    for e in entities:
        if e.has_preferred_sense():
            preferred_sense = e.get_preferred_sense()
            if e.kb.get_ent_type(preferred_sense) == "artwork":
                artist = e.kb.get_data_for(preferred_sense, "ARTIST")
                if e.source.lower() not in MOST_FAMOUS_ARTWORKS:
                    if artist not in global_artist_names:
                        e.set_preferred_sense(None)

    # resolving coreferences
    name_coreferences = [e for e in entities if e.source not in PRONOUNS and not e.source.lower().startswith("the")]
    resolve_coreferences(name_coreferences, context, print_all, register)
    resolve_coreferences(entities, context, print_all, register)

    # resolving overlapping entities and proper nouns
    entity_offsets = set()
    for e in entities:
        entity_offsets.update(set(range(e.start_offset, e.end_offset + 1)))
    proper_nouns = find_proper_nouns(original_input)
    proper_nouns = [pn for pn in proper_nouns if set(range(pn[0], pn[1] + 1)) - entity_offsets != set()]
    proper_noun_offsets = set()
    for pn in proper_nouns:
        proper_noun_offsets.update(set(range(pn[0], pn[1] + 1)))
    new_entities = []
    for e in entities:
        if not (e.start_offset in proper_noun_offsets and e.end_offset in proper_noun_offsets and (e.start_offset, e.end_offset) not in proper_nouns):
            new_entities.append(e)
    entities = new_entities

    # determining whether two entities lie next to each other
    input_string_decoded = input_string.decode("utf8", "replace")
    ent_ind = 0
    for ent_ind in range(1, len(entities)):
        ent = entities[ent_ind]
        ent_bef = entities[ent_ind - 1]
        if ent.has_preferred_sense() and ent.source.lower() not in PRONOUNS:
            if ent_bef.has_preferred_sense() and ent_bef.source.lower() not in PRONOUNS:
                # if both entities are divided only by a space
                if re.match("^ $", input_string_decoded[ent_bef.end_offset:ent.start_offset]):
                    ent.next_to_same_type = True
                    ent_bef.next_to_same_type = True

    # filtering out entities that are next to another entity (excluding dates)
    entities = [e for e in entities if not e.next_to_same_type]

    # updating entities_and_dates
    entities_and_dates = [e for e in entities_and_dates if isinstance(e, dates.Date) or e in entities]

    # omitting entities whithout a sense
    if not print_all:
        entities_and_dates = [e for e in entities_and_dates if isinstance(e, dates.Date) or e.has_preferred_sense()]
    else:
        for e in entities_and_dates:
            if isinstance(e, Entity):
                e.set_preferred_sense(None)
        entities_and_dates = [e for e in entities_and_dates if isinstance(e, dates.Date) or (e.is_coreference and e.partial_match_senses) or (not e.is_coreference and e.senses)]

    for ent in entities_and_dates:
        ent.display_score = print_score

    if print_result:
        print("\n".join(map(str, entities_and_dates)))

    return entities_and_dates

if __name__ == "__main__":

    # argument parsing
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-a', '--all', action='store_true', default=False, dest='all', help='Prints all entities without disambiguation.')
    group.add_argument('-s', '--score', action='store_true', default=False, dest='score', help='Prints all possible senses with respective score values.')
    parser.add_argument('-d', '--daemon-mode', action='store_true', default=False, help='Runs ner.py in daemon mode.')
    parser.add_argument('-f', '--file',  help='Uses a given file as as an input.')
    parser.add_argument('-r', '--remove-accent', action='store_true', default=False, help="Removes accent in input.")
    parser.add_argument('-l', '--lowercase', action='store_true', default=False, help="Changes all characters in input to the lowercase characters.")

    arguments = parser.parse_args()
    # allowed tokens for daemon mode
    tokens = set(["NER_NEW_FILE", "NER_END", "NER_NEW_FILE_ALL", "NER_END_ALL", "NER_NEW_FILE_SCORE", "NER_END_SCORE"])

    # loading knowledge base
    kb = ner_knowledge_base.KnowledgeBase()

    try:
        kb.start()
        kb.initName_dict()

        if arguments.daemon_mode:
            input_string = ""
            while True:
                line = sys.stdin.readline().rstrip()
                if line in tokens:
                    if "ALL" in line:
                        recognize(kb, input_string, print_all=True, lowercase=arguments.lowercase, remove=arguments.remove_accent)
                    elif "SCORE" in line:
                        recognize(kb, input_string, print_score=True, lowercase=arguments.lowercase, remove=arguments.remove_accent)
                    else:
                        recognize(kb, input_string, print_all=False, lowercase=arguments.lowercase, remove=arguments.remove_accent)
                    print(line)
                    sys.stdout.flush()
                    input_string = ""
                    if "END" in line:
                        break
                else:
                    input_string += line + "\n"
        else:
            # reading input data from file
            if arguments.file:
                with open(arguments.file) as f:
                    input_string = f.read()
            # reading input data from stdin
            else:
                input_string = sys.stdin.read()
            input_string = input_string.strip()
            recognize(kb, input_string, arguments.all, print_score=arguments.score, lowercase=arguments.lowercase, remove=arguments.remove_accent)
    finally:
        pass
#        kb.end()
